/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package pp4u;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetSocketAddress;
import java.net.Socket;
import pp4u_packet.Packet;
import pp4u_packet.PacketGet;
import pp4u_packet.PacketPing;
import pp4u_packet.PacketPong;
import pp4u_packet.PacketPut;

/**
 * 受信タスク
 * @author tondol
 */
public class TaskDownload extends TaskAbstract implements Runnable {

    private Socket socket = null;
    private ObjectInputStream in = null;
    private ObjectOutputStream out = null;
    private InfoNode gaddr = null;
    private InfoNode paddr = null;
    private Thread thread = null;
    private InfoKey key = null;
    private volatile long current = 0;
    private volatile long length = 0;
    private final int blockLength = 2*1024*1024;

    /**
     * キー情報のみ設定して生成
     */
    public TaskDownload(InfoKey key) {
        this.key = key;
        thread = new Thread(this);
        thread.start();
    }

    public void stop() {
        try {
            if (socket != null) {
                socket.close();
            }
        } catch (IOException iex) {
        } finally {
            socket = null;
            thread = null;
        }
    }

    /**
     * スレッド本体
     */
    public void run() {
        FileOutputStream fst = null;
        BufferedOutputStream ost = null;

        try {
            // ソケット接続 - タイムアウト5秒
            String hostname = key.getNode().getHostname();
            int port = key.getNode().getPort();
            InetSocketAddress address = new InetSocketAddress(hostname, port);
            this.socket = new Socket();
            this.socket.connect(address, 5*1000);
            // PING送信，PONG受信
            this.out = new ObjectOutputStream(socket.getOutputStream());
            PacketPing ping = new PacketPing(PacketPing.TRANSFER);
            send(ping);
            this.in = new ObjectInputStream(socket.getInputStream());
            PacketPong pong = (PacketPong) recv();
            // ホスト情報を取得
            String ghost = socket.getInetAddress().getHostAddress();
            int gport = pong.getAddress().getPort();
            this.paddr = pong.getAddress();
            this.gaddr = new InfoNode(ghost, gport);
            
            // ファイルを開く
            File file = new File(key.getDownloadPath());
            fst = new FileOutputStream(file, true);
            ost = new BufferedOutputStream(fst);
            // 情報取得
            this.length = key.getSize();
            this.current = file.length();
            
            // 受信完了まで繰り返す
            while (length - current > 0) {
                // GET送信とPUT受信
                PacketGet get = new PacketGet(key, current, blockLength);
                send(get);
                PacketPut put = (PacketPut) recv();
                // ブロックを書き込む
                int loaded = (int) Math.min(length - current, blockLength);
                ost.write(put.getBytes(), 0, loaded);
                current += loaded;
            }
        } catch (Exception ex) {
            AppMain.debug("# 受信タスクエラー");
            AppMain.debug("@ " + ex);
        } finally {
            try {
                if (ost != null) {
                    ost.close();
                }
            } catch (IOException iex) {
            } finally {
                stop();
            }
        }
    }

    private void send(Packet packet) throws IOException {
        out.writeObject(packet);
        out.reset();
    }

    private Packet recv() throws IOException, ClassNotFoundException {
        return (Packet) in.readObject();
    }

    //////////////////////////////////////////////
    // 取得・設定メソッド
    //////////////////////////////////////////////
    @Override
    public boolean isAlive() {
        if (socket == null) {
            return false;
        } else {
            return true;
        }
    }

    @Override
    public String getType() {
        return "受信タスク";
    }

    @Override
    public String getFilename() {
        if (key == null) {
            return null;
        } else {
            return key.getFilename();
        }
    }

    @Override
    public InfoNode getNode() {
        if (key == null) {
            return null;
        } else {
            return key.getNode();
        }
    }

    @Override
    public String getStatus() {
        int per = (int) (100.0 * (double) current / (double) length);
        return per + "% @ " + current + "/" + length;
    }

    //////////////////////////////////////////////
    // インターフェイスの実装
    //////////////////////////////////////////////
    @Override
    public boolean equals(Object obj) {
        if (obj == null || getClass() != obj.getClass()) {
            return false;
        }
        final TaskDownload other = (TaskDownload) obj;
        if (!this.key.equals(other.key)) {
            return false;
        }
        return true;
    }

    @Override
    public int hashCode() {
        return key.hashCode();
    }
}
